Sys.setlocale("LC_ALL", 'en_GB.UTF-8')
Sys.setenv(LANG = "en_US.UTF-8")

getwd()
setwd("~/Dropbox/Milan/2021 Data Access/Class 5 - Week 1 Friday/")   # Set this to the folder on your computer where you've saved the files for this class.

install.packages("RMariaDB")
install.packages("jsonlite")

# In today's class, we're primarily going to control MySQL directly from R.

# It's possible to control the database from an application like MySQL Workbench or Sequel Pro, and that can be a useful
# way to quickly run SQL commands or check that things are working properly.

# However, for data cleaning, storage and analysis, it's usually more effective to control the database from your programming
# language - whether that's R, Python or something else. 




#-----------------------------------------------------------------------------#
# 1. Connecting to a Database from R
#-----------------------------------------------------------------------------#

# The DBI library is the standard way to interact with a database in R. 
# A useful thing about it is that it has various "drivers" for different SQL databases - in this case we're using RMariaDB, which is
# the driver for both MySQL and another database called MariaDB. However, even if you changed to using a totally different database,
# the DBI commands you'd use in R would stay exactly the same! You'd just change the details of the "dbConnect" command
# to connect to a different type of database, and after that, everything would be done in precisely the same way.

library(DBI)

# This command creates an object called "con", which is R's connection to the database.
# It's not like a regular object in R (which would contain data of some kind) - it's more like an abstract way to 
# refer to the connection itself. You can't look inside it and see data, but otherwise you can treat it like
# any other variable or object.

# Make sure you set "username" and "password" to the ones you selected when you installed MySQL.

con <- dbConnect(RMariaDB::MariaDB(),
                 host = "localhost",
                 username = "dapsco",
                 password = "daps2021")


# I'll mention this again at the end, but every time you OPEN a connection to a database, you should also remember to 
# CLOSE it when you're done. The command to do that is:

# dbDisconnect(con)

# Usually, failing to disconnect doesn't cause any huge problems when you're just working with MySQL on your own 
# computer. But if you have multiple users accessing the same database (like one stored on a server or in the Cloud),
# and you open a bunch of connections without closing any of them, then your colleagues might not be able to 
# access the data because you've hogged all the available connections.




#-----------------------------------------------------------------------------#
# 2. Executing SQL Commands and Creating Databases
#-----------------------------------------------------------------------------#

# Since this is a brand new installation of MySQL, we first want to create a database for us to work in.

# Okay, wait - isn't MySQL the database? 
# Actually, MySQL is a database server. It can contain many databases. Usually, you'll create a new one
# for each project you work on. Moreover, each database can contain many different tables (we'll create
# a table in a moment).

# Here we're going to use the "dbExecute" command to run a line of SQL code in MySQL. It's a very simple 
# line of code that will just create a new, blank database for us to work in.

# Today we're going to be using a dataset made of customer reviews from Amazon.com, so let's call the
# database something appropriate.

dbExecute(con, "CREATE DATABASE amazon_reviews")

# (NOTE: If you run this command a second time, it will cause an error. Why? Because you already have
# a database called "amazon_reviews", so trying to create the same database again will give you an error.)


# Next, let's run another command to list the databases available on our MySQL server.

# This time, we use a different R command: "dbSendStatement" instead of "dbExecute".
# The difference between these commands is that "dbExecute" runs your SQL code and just returns a single
# number (indicating how many rows of data your command affected). "dbSendStatement" instead returns a
# new object with the results of your command - which could be rows of a data table, or some other
# values you've asked MySQL to show you.

result <- dbSendStatement(con, "SHOW DATABASES")
dbFetch(result, n = -1)
dbClearResult(result)

# "dbFetch" is a command that you use to fetch data from that "results" object. A useful thing about the
# results object is that until you "fetch" from it, it doesn't actually contain any data; like the "con"
# object, it's more like an abstract representation of a connection to some data that exists in MySQL.

# Why is that useful? Well, here we've fetched all of the results at once (by using "n = -1", which is
# how we tell dbFetch to give you everything), but you could also fetch results in batches. For example,
# you could set "n = 100" which will give you a hundred records; then run dbFetch again to get the next 
# hundred records... And so on. In this way, you can process HUGE datasets that are much too big to be
# handled by R as a single dataframe, by accessing them one chunk of data at a time.

# Note that along with the amazon_reviews database that we created, MySQL has a few other databases in it.
# Don't worry about these; they're basically used by the MySQL server to store its own settings and 
# information about the data it contains. Generally you shouldn't need to do anything to these, and you 
# definitely shouldn't delete them, as doing so could break your database!

# Okay, one last thing in this section - let's tell MySQL that we want to use the amazon_reviews database
# from now on. Since this command doesn't return any data, we use dbExecute for it.

dbExecute(con, "USE amazon_reviews")



#-----------------------------------------------------------------------------#
# 3. Today's Example Data - JSON Data Importing
#-----------------------------------------------------------------------------#

# The data we're going to use today for populating a MySQL database is coming from a type of file you
# may not have seen before, namely a JSON file. We mentioned this briefly in an earlier class, and 
# if you've scraped data from an API on the web you've likely seen something very similar.

# Basically, this is a file full of data stored in a format like this:

#{"reviewerID": "A14VAT5EAX3D9S", 
# "asin": "1384719342", 
# "reviewerName": "Jake", 
# "helpful": [13, 14], 
# "reviewText": "The product does exactly as it should and is quite affordable.I did not realized it was double screened 
#                until it arrived, so it was even better than I had expected.As an added bonus, one of the screens carries 
#                a small hint of the smell of an old grape candy I used to buy, so for reminiscent's sake, I cannot stop 
#                putting the pop filter next to my nose and smelling it after recording. :DIf you needed a pop filter, this 
#                will work just as well as the expensive ones, and it may even come with a pleasing aroma like mine did!Buy 
#                this product! :]", 
# "overall": 5.0, 
# "summary": "Jake", 
# "unixReviewTime": 1363392000, 
# "reviewTime": "03 16, 2013"}

# This kind of file has become a popular alternative to CSV, especially since it's the usual format for sending data
# around using internet APIs. To load this data into R, there's a handy library called "jsonlite".

library(jsonlite)

# jsonlite does have a simple "read_json" function, but that only works when your file contains a single JSON object. Our 
# file of Amazon reviews contains a lot of small individual objects (one per review, each enclosed in curly brackets {}), 
# so we use a different function, "stream_in" - which is designed to read a "stream" of individual objects.

reviews <- stream_in(file('amazon_reviews.json'))

View(reviews)

# Okay, before we put this data into MySQL, we need to do some tidying. Specifically, that reviewTime column is a 
# strangely formatted date, so we'll use the lubridate commands we looked at yesterday to get it into shape.

# We also don't actually need to keep all of those columns - for example, unixReviewTime is just a different way of 
# recording the reviewTime, so we don't need them both.

# I would suggest that we should keep the reviewerID and reviewerName, the 'asin' (this is the Amazon product code for the 
# item being reviewed), the overall score, review summary, review text, and review time.

library(dplyr)

tidy_reviews <- reviews %>%
  mutate(reviewTime = lubridate::mdy(reviewTime)) %>%
  select(-helpful, -unixReviewTime)

View(tidy_reviews)

#-----------------------------------------------------------------------------#
# 4. Designing a MySQL Table
#-----------------------------------------------------------------------------#


# Okay, our data is tidy and ready to go into the database. But first... Remember that to use MySQL (or any SQL 
# database), you need to design the table your data will go into before you can actually insert any data.

# To make a table in MySQL, we use the CREATE TABLE command. Set the name of the table, and then in brackets, 
# describe each column you want to put in the table - its name, the type of data it will hold, and any other parameters 
# or information about the column. Finally, you select the PRIMARY KEY for the table - a unique value, usually the
# auto-generated "id" column, that will identify each row - and you're ready to create your table.

# MySQL Column Types

# The most common types of column in SQL are "varchar" - which is a fixed-width column of text, whose maximum size you 
# need to specify - "int", for integers, "decimal" for decimal numbers, "date" and "datetime" for dates and times, and "text" 
# for large amounts of text - up to 65,000 characters, which is usually plenty for even a large newspaper article.

# There are other types of column as well - columns for storing JSON data directly, or for storing binary data like 
# images, for example - but these are the ones you're most likely to encounter.

# For this class, we're going to create our table using an SQL command, as shown below - but if you're quickly designing
# data tables for a project, it's also possible to create them in the editor provided by MySQL Workbench That can be a
# lot more intuitive than having to write out this long command (but it helps to understand what's going on in the 
# background, too).

dbExecute(con, "CREATE TABLE `reviews` (
  `id` int unsigned NOT NULL AUTO_INCREMENT,
  `reviewerID` varchar(30),
  `reviewerName` varchar(50),
  `productID` varchar(30),
  `stars` int,
  `summaryText` varchar(200),
  `reviewText` text,
  `postedDate` date,
  PRIMARY KEY (`id`)
  );")

#-----------------------------------------------------------------------------#
# 5. Inserting Data into MySQL
#-----------------------------------------------------------------------------#

# Now we have a dataframe with our clean data in it, and a newly-created MySQL table to hold it. It's time to learn about 
# the INSERT command in MySQL, so we can move that data into the database.

# The syntax for the INSERT command is thankfully very simple - it works as follows:

# INSERT INTO my_table (column1, column2, column3, ...) VALUES (value1, value2, value3, ...)

# The only tricky part is making sure that the values are in the same order as the columns, so you're not trying to insert 
# data into the wrong places!

# With R's DBI system, you can put question marks (?) in place of values in the SQL command, and then allow DBI to fill in
# those values from a list of parameters, "params", which you include in the dbExecute or dbSendStatement command.

# That's useful in many instances, but especially when you actually want to insert or update a lot of data - for example,
# putting the entire content of a dataframe in R into your database, as we're doing here. By putting large lists or columns of data 
# in place of the question marks, DBI knows that we want to insert a lot of data into the table with this single command. 

dbExecute(con, 
          "INSERT INTO reviews (reviewerID, reviewerName, productID, stars, summaryText, reviewText, postedDate) VALUES (?, ?, ?, ?, ?, ?, ?)",
          params = list(tidy_reviews$reviewerID, 
                        tidy_reviews$reviewerName,
                        tidy_reviews$asin,
                        tidy_reviews$overall,
                        tidy_reviews$summary, 
                        tidy_reviews$reviewText,
                        tidy_reviews$reviewTime))

# An important thing to know about this is that you can do INSERT INTO commands on a table in SQL as many times as you like.
# For example, if you're downloading data from an API, you can INSERT it into the database as you receive it, and it'll just
# add it to the existing data. 
# This is also useful if you have data that came in multiple CSV or Excel files (e.g. one file per month, or per year) - you can
# process the files one at a time and add them all to the same table in SQL for use in later analysis.



#-----------------------------------------------------------------------------#
# 5.1 Another way to Create and Populate SQL Tables
#-----------------------------------------------------------------------------#

# There is another way of creating a table and inserting data which doesn't involve writing any SQL commands
# at all - it's faster and simpler, but the downside is that you don't have as much control over your
# SQL table, and it's possible to run into problems as a result.

# DBI has a command called "dbCreateTable", which will automatically create a new SQL table based on the
# columns of an existing dataframe. Then, you can add the data from that dataframe (or another dataframe with 
# identical column names) to SQL using the command "dbAppendTable".

# In our case, with the table tidy_reviews, the commands would be:

# dbCreateTable(con, "reviews", tidy_reviews)
# dbAppendTable(con, "reviews", tidy_reviews)

# Much simpler of course - but you have less control, and it's very important to ensure that your dataframe
# column names are correct and legible!


#-----------------------------------------------------------------------------#
# 6. Selecting Data from SQL
#-----------------------------------------------------------------------------#


# Now all of our reviews are safely stored in MySQL. If you just want to take a look at them, try opening MySQL 
# Workbench and clicking on the "Schemas" tab (a schema is another name for a database in MySQL). You should see 
# the amazon_reviews database, and inside it, under Tables, the reviews table. If you view that table (there's a 
# pop-up icon to look at it directly), you can see all our data neatly stored in there.

# MySQL Workbench is really useful for quickly checking things in your database - you can also use it to try out 
# writing SQL statements, since you can run them directly from that window. Very often, it's a good idea to try out 
# your SQL in Workbench first before using it in your R code, since you can see the results and edit the code more 
# quickly and easily.


# Next, let's look at some different ways to filter and select this data in R. Getting data out of a database is done 
# with a SELECT statement - let's start with a simple example, where we select every review in the second half of 
# 2012 that gave less than three stars.

result <- dbSendStatement(con, "SELECT * FROM reviews 
                                WHERE postedDate BETWEEN '2012-07-01' AND '2012-12-31' 
                                  AND stars < 3")
subset <- dbFetch(result, n = -1)
dbClearResult(result)
View(subset)

# SELECT statements can be very simple - the most simple of all is just SELECT * FROM table, which will select 
# everything in a database table. Adding WHERE clauses filters the results by matching against different variables.

# In the next example, we'll introduce a few additional things you can do in a SELECT command:

# COUNT() will count the number of rows matching your filter.

# GROUP BY will collect together all of the rows where a certain column has the exact same value. 
#    This can be powerful especially when you use it alongside other commands like COUNT(), because it lets you see 
#    how many rows were grouped together.

# ORDER BY will order your results by the column you specify. 
#    You can add ASC (ascending) or DESC (descending) to change the direction of the order.

# Finally, when handling dates and times, you can extract just one part of the date by using YEAR(), MONTH() or DAY().

# Using all of these new elements, let's find out what the distribution of reviews in our database is over the years.

result <- dbSendStatement(con, "SELECT YEAR(postedDate), COUNT(*) FROM reviews 
                                 GROUP BY YEAR(postedDate) 
                                 ORDER BY YEAR(postedDate) ASC")
annual_stats <- dbFetch(result, n = -1)
dbClearResult(result)
show(annual_stats)

# We could also use much the same approach to figure out who the most prolific reviewers in the data set are...

result <- dbSendStatement(con, "SELECT reviewerID, reviewerName, COUNT(reviewerID) FROM reviews 
                                 GROUP BY reviewerID 
                                 ORDER BY COUNT(reviewerID) DESC")
reviewer_stats <- dbFetch(result, n = 10)   # We only want to see the top ten most prolific reviewers
dbClearResult(result)
show(reviewer_stats)


#-----------------------------------------------------------------------------#
# 7. In-Class Exercises (Time Permitting)
#-----------------------------------------------------------------------------#

# 1) What are the most-reviewed products in the data set?
#  
#  Using the same kind of SQL queries we used above, can you figure out what the most commonly reviewed products in the data set were?
  



# 2) Is there a seasonal pattern to reviews?
#  
# All of these reviews are from the musical equipment and instruments section of Amazon. Since that kind of thing is 
# often given as a Christmas present, I wonder if there are more reviews in December and January than at other times of 
# the year? Can you find a seasonal pattern in the data using SQL?
  



# 3) Find the number of reviews and average star ratings for the products.

# When you visit a product page on Amazon, it tells you two things about the reviews: the number of times the product 
# has been reviewed, and the average star rating from those reviews. Try to get that data from a single SQL command.

# (Hint: There is an SQL function AVG() which gets the average value of a column. 
# It works pretty much the same as COUNT(), which you've seen already.)



# And remember... Always disconnect at the end of your SQL session!
dbDisconnect(con)
